"""This file contains shared fixtures and pytest hooks.

https://docs.pytest.org/en/6.2.x/fixture.html#conftest-py-sharing-fixtures-across-multiple-files
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol

from pytest import fixture

from greatday.__main__ import main as gtd_main


if TYPE_CHECKING:  # fixes pytest warning
    from clack.pytest_plugin import MakeConfigFile


pytest_plugins = ["clack.pytest_plugin"]


class MainType(Protocol):
    """Type returned by main() fixture."""

    def __call__(self, *args: str, **kwargs: Any) -> int:
        """The signature of the main() function."""


@fixture
def main(make_config_file: MakeConfigFile, tmp_path: Path) -> MainType:
    """Returns a wrapper around greatday's main() function."""

    data_dir = tmp_path / "data"

    def inner_main(*args: str, **kwargs: Any) -> int:
        if "data_dir" not in kwargs:
            kwargs["data_dir"] = data_dir

        cfg_kwargs = {k: str(v) for (k, v) in kwargs.items()}

        config_file = make_config_file("greatday_test_config", **cfg_kwargs)
        argv = ["greatday", "-c", str(config_file.path)] + list(args)
        return gtd_main(argv)

    return inner_main



"""Test for greatday's 'add' subcommand."""

from __future__ import annotations

from .conftest import MainType


def test_add(main: MainType) -> None:
    """Tests the 'add' subcommand."""
    assert main("add", "foo") == 0



"""Editor (e.g. vim) code lives here."""

from __future__ import annotations

from vimala import vim

from ._session import GreatSession


def edit_and_commit_todos(session: GreatSession) -> None:
    """Edit and commit todo changes to disk."""
    old_todos = list(session.repo.todo_group)
    if not old_todos:
        return

    vim(session.path).unwrap()

    for otodo in old_todos:
        key = otodo.ident

        new_todo = session.repo.get(key).unwrap()
        if otodo != new_todo:
            break
    else:
        if len(old_todos) == len(session.repo.todo_group):
            return

    session.commit()



"""Common code used throughout this package."""

from __future__ import annotations

import datetime as dt
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Final

from dateutil.relativedelta import relativedelta
from typist import PathLike


if TYPE_CHECKING:
    from ._todo import GreatTodo


CTX_TODAY: Final = "today"

MONDAY: Final = 0
TUESDAY: Final = 1
WEDNESDAY: Final = 2
THURSDAY: Final = 3
FRIDAY: Final = 4
SATURDAY: Final = 5
SUNDAY: Final = 6


def drop_word_from_desc(
    desc: str,
    bad_word: str,
    *,
    op: Callable[[str, str], bool] = lambda x, y: x == y,
) -> str:
    """Removes `bad_word` from the todo description `desc`."""
    desc_words = desc.split(" ")
    new_desc_words = []
    for word in desc_words:
        if not op(word, bad_word):
            new_desc_words.append(word)
    return " ".join(new_desc_words)


def get_relative_date(spec: str, *, start_date: dt.date = None) -> dt.date:
    """Converts `spec` to a timedelta and adds it to `date`.

    Args:
        spec: A timedelta specification string (e.g. '1d', '2m', '3y',
          'weekdays').
        start_date: The return value is a function of this argument and the
          timedelta constructed from `spec`. Defaults to today's date.

    Examples:
        # Imports
        >>> import datetime as dt

        # Helper Functions
        >>> to_date = lambda x: dt.datetime.strptime(x, "%Y-%m-%d")
        >>> from_date = lambda x: x.strftime("%Y-%m-%d")
        >>> grd = lambda x, y: from_date(
        ...   get_relative_date(x, start_date=to_date(y))
        ... )

        # Default start date.
        >>> D = "2000-01-31"

        # Tests
        >>> grd("7d", D)
        '2000-02-07'

        >>> grd("7D", D)
        '2000-02-07'

        >>> grd("1m", D)
        '2000-02-29'

        >>> grd("1m", "2001-01-31")
        '2001-02-28'

        >>> grd("2M", D)
        '2000-03-31'

        >>> grd("3m", D)
        '2000-04-30'

        >>> grd("20y", D)
        '2020-01-31'

        >>> grd("weekdays", "2022-02-11")
        '2022-02-14'
    """
    spec = spec.lower()
    if start_date is None:
        start_date = dt.date.today()

    delta: dt.timedelta | relativedelta
    if spec == "weekdays":
        weekday = start_date.weekday()
        days = {FRIDAY: 3, SATURDAY: 2}.get(weekday, 1)
        delta = dt.timedelta(days=days)
    else:
        ch = spec[-1]
        N = int(spec[:-1])

        if ch == "d":
            delta = dt.timedelta(days=N)
        elif ch == "m":
            delta = relativedelta(months=N)
        else:
            assert ch == "y"
            delta = relativedelta(years=N)

    return start_date + delta


def is_tickler(todo: GreatTodo) -> bool:
    """Returns True iff `todo` is a tickler todo."""
    return bool(todo.metadata.get("tickle", None) is not None)


def dt_from_date_and_hhmm(date: dt.date, hhmm: str) -> dt.datetime:
    """Given a date and a string of the form HHMM, construct a datetime."""
    spec = f"{date.year}-{date.month}-{date.day} {hhmm}"
    result = dt.datetime.strptime(spec, "%Y-%m-%d %H%M")
    return result


def matches_date_fmt(date_spec: str) -> bool:
    """Returns True iff date_spec matches the magodo date format.."""
    return len(date_spec) == 10 and date_spec.count("-") == 2


def init_yyyymm_path(root: PathLike, *, date: dt.date = None) -> Path:
    """Returns a Path of the form /path/to/root/YYYY/MM.txt.

    NOTE: Creates the /path/to/root/YYYY directory if necessary.
    """
    root = Path(root)
    if date is None:
        date = dt.date.today()

    year = date.year
    month = date.month

    result = root / str(year) / f"{month:0>2}.txt"
    result.parent.mkdir(parents=True, exist_ok=True)
    return result



"""Functions / classes used to create greatday's TUI."""

from __future__ import annotations

from dataclasses import dataclass
import datetime as dt
from functools import partial
from typing import Any

import magodo
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text
from textual.app import App
from textual.reactive import Reactive
from textual.widgets import Footer, Header, Static
from textual_inputs import TextInput
from typist import PathLike

from ._editor import edit_and_commit_todos
from ._repo import GreatRepo
from ._session import GreatSession
from ._tag import Tag


class GreatHeader(Header):
    """Override the default Header for Styling"""

    def __init__(self) -> None:
        super().__init__()
        self.tall = False
        self.style = Style(color="white", bgcolor="rgb(98,98,98)")

    def render(self) -> Table:
        """Returns renderable header."""
        header_table = Table.grid(padding=(0, 1), expand=True)
        header_table.add_column(justify="left", ratio=0, width=8)
        header_table.add_column("title", justify="center", ratio=1)
        header_table.add_column("clock", justify="right", width=8)
        header_table.add_row(
            "🔤", self.full_title, self.get_clock() if self.clock else ""
        )
        return header_table


class GreatFooter(Footer):
    """Override the default Footer for Styling"""

    def make_key_text(self) -> Text:
        """Create text containing all the keys."""
        text = Text(
            style="white on rgb(98,98,98)",
            no_wrap=True,
            overflow="ellipsis",
            justify="left",
            end="",
        )
        for binding in self.app.bindings.shown_keys:
            key_display = (
                binding.key.upper()
                if binding.key_display is None
                else binding.key_display
            )
            hovered = self.highlight_key == binding.key
            key_text = Text.assemble(
                (
                    f" {key_display} ",
                    "reverse" if hovered else "default on default",
                ),
                f" {binding.description} ",
                meta={
                    "@click": f"app.press('{binding.key}')",
                    "key": binding.key,
                },
            )
            text.append_text(key_text)
        return text


class StatsWidget(Static):
    """Widget that shows Todo statistics."""

    repo: Reactive[GreatRepo | None] = Reactive(None)

    def __init__(self, repo: GreatRepo, *args: Any, **kwargs: Any) -> None:
        super().__init__("", *args, **kwargs)
        self.repo = repo

    def render(self) -> Panel:
        """Render the statistics widget."""
        assert self.repo is not None

        tag = Tag.from_query("@inbox done=0")
        inbox_todos = self.repo.get_by_tag(tag).unwrap()
        inbox_count = len(inbox_todos) if inbox_todos else 0

        today = dt.date.today()
        tag = Tag.from_query(
            "tickle<={0} snooze?<={0} done=0".format(magodo.from_date(today))
        )
        tickler_todos = self.repo.get_by_tag(tag).unwrap()
        tickler_count = len(tickler_todos) if tickler_todos else 0

        tag = Tag.from_query("@today done=0")
        today_todos = self.repo.get_by_tag(tag).unwrap()
        today_count = len(today_todos) if today_todos else 0

        return Panel(
            f"INBOX: {inbox_count}\nTICKLERS: {tickler_count}\nTODAY:"
            f" {today_count}\n",
            title="Statistics",
        )


@dataclass
class Context:
    """Mutable TUI Context.

    Used to preserve state after closing the GreatApp instance, running vim,
    and then opening a new GreatApp instance.

    Attributes:
        query: The active todo query string.
        is_user_query: Is this query one the user selected or just the
          default?
        edit_todos: After closing the TUI, should we open up vim to edit
          matching todos?
    """

    query: str
    is_user_query: bool = False
    edit_todos: bool = False


class GreatApp(App):
    """Textual TUI Application Class."""

    def __init__(
        self, *, repo: GreatRepo, ctx: Context, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)

        self.repo = repo
        self.ctx = ctx

        text_input = partial(TextInput, name="input")
        if ctx.is_user_query:
            self.input_widget = text_input(value=self.ctx.query)
        else:
            self.input_widget = text_input(placeholder=self.ctx.query)

        cursor = (
            "|",
            Style(
                color="black",
                blink=True,
                bold=True,
            ),
        )
        self.input_widget.cursor = cursor

        self.main_widget = Static(
            Panel(
                _todo_lines_from_query(self.repo, self.ctx.query),
                title="Main",
            ),
            name="main",
        )
        self.stats_widget = StatsWidget(self.repo)

    async def on_load(self) -> None:
        """Configure key bindings."""
        await self.bind("escape", "change_mode('normal')", "Normal Mode")
        await self.bind("enter", "submit", "Submit")
        await self.bind("e", "edit", "Edit Todos")
        await self.bind("i", "change_mode('insert')", "Insert Mode")
        await self.bind("q", "quit", "Quit")

    async def on_mount(self) -> None:
        """Configure layout."""
        # configure header and footer...
        await self.view.dock(GreatHeader(), edge="top")
        await self.view.dock(GreatFooter(), edge="bottom")

        # configure other widgets...
        await self.view.dock(self.input_widget, edge="bottom", size=10)
        await self.view.dock(self.stats_widget, edge="left", size=50)
        await self.view.dock(self.main_widget, edge="top")

    async def action_change_mode(self, mode: str) -> None:
        """Action to toggle to/from insert mode and other modes."""
        if mode == "insert":
            self.input_widget.title = "INSERT"
            self.input_widget.refresh()
            await self.input_widget.focus()
        elif mode == "normal":
            self.input_widget.title = ""
            self.input_widget.refresh()
            await self.main_widget.focus()
        else:
            raise AssertionError(f"Bad mode: {mode!r}")

    async def action_submit(self) -> None:
        """Called when the user hits <Enter>."""
        self.ctx.query = self.input_widget.value
        self.ctx.is_user_query = True
        self.input_widget.placeholder = ""
        text = _todo_lines_from_query(self.repo, self.ctx.query)
        await self.main_widget.update(Panel(text, title="Todo List"))
        await self.action_change_mode("normal")

    async def action_edit(self) -> None:
        """Edits todos which match the current todo query."""
        self.ctx.edit_todos = True
        await self.action_quit()


def _todo_lines_from_query(repo: GreatRepo, query: str) -> str:
    tag = Tag.from_query(query)
    todos = repo.get_by_tag(tag).unwrap()

    result = ""
    for todo in sorted(todos):
        result += todo.to_line() + "\n"

    return result


def start_textual_app(data_dir: PathLike, repo_path: PathLike) -> None:
    """Starts the TUI using the GreatApp class."""
    repo = GreatRepo(data_dir, repo_path)
    ctx = Context("@today")
    run_app = partial(
        GreatApp.run,
        repo=repo,
        ctx=ctx,
        title="Greatday TUI",
        log="greatday_textual.log",
    )
    run_app()

    while ctx.edit_todos:
        tag = Tag.from_query(ctx.query)
        with GreatSession(data_dir, repo_path, tag) as session:
            edit_and_commit_todos(session)

        ctx.edit_todos = False
        run_app()



"""Contains the greatday package's main entry point."""

from __future__ import annotations

import clack

from . import APP_NAME
from ._config import clack_parser
from ._runners import ALL_RUNNERS


main = clack.main_factory(APP_NAME, runners=ALL_RUNNERS, parser=clack_parser)
if __name__ == "__main__":
    main()



"""Contains the Tag class."""

from __future__ import annotations

from dataclasses import dataclass
import datetime as dt
import operator
from typing import Any, Callable, Iterable

from logrus import Logger
import magodo
from magodo import MetadataCheck
from magodo.types import MetadataFunc, Priority

from ._common import matches_date_fmt


logger = Logger(__name__)


@dataclass(frozen=True)
class Tag:
    """Tag used to filter Todos."""

    contexts: Iterable[str] = ()
    create_date: dt.date | None = None
    done_date: dt.date | None = None
    done: bool | None = None
    epics: Iterable[str] = ()
    metadata_checks: Iterable[MetadataCheck] = ()
    priorities: Iterable[Priority] = ()
    projects: Iterable[str] = ()

    @classmethod
    def from_query(cls, query: str) -> Tag:
        """Build a Tag using a query string."""
        contexts: list[str] = []
        create_and_done: list[dt.date | None] = [None, None]
        done: bool | None = None
        epics: list[str] = []
        metadata_checks: list[MetadataCheck] = []
        priorities: list[Priority] = []
        projects: list[str] = []

        for word in query.split(" "):
            # ----- Regular Tags (i.e. projects, epics, and contexts)
            is_a_tag = True
            for prefix, prop_list in [
                ("@", contexts),
                ("#", epics),
                ("+", projects),
            ]:
                if word.startswith(prefix):
                    logger.debug("Filter on property.", word=word)
                    prop_list.append(word[1:])
                    break

                if word.startswith(f"!{prefix}"):
                    logger.debug("Filter on negative property.", word=word)
                    prop_list.append(f"-{word[2:]}")
                    break
            else:
                is_a_tag = False

            if is_a_tag:
                continue

            # ----- Create and Done Dates
            is_date_range = False
            for i, date_prefix in enumerate(["create=", "done="]):
                if word.startswith(date_prefix):
                    date_spec = word[len(date_prefix) :]
                    if not matches_date_fmt(date_spec):
                        logger.debug(
                            "Date does not match required date format.",
                            date_spec=date_spec,
                        )
                        continue

                    create_and_done[i] = magodo.to_date(date_spec)
                    logger.debug(
                        "Filter on date.",
                        prefix=date_prefix,
                        date=create_and_done[i],
                    )
                    is_date_range = True

            if is_date_range:
                continue

            # ----- Is open todo or done todo?
            done_prefix = "done="
            if word.startswith(done_prefix):
                zero_or_one = word[len(done_prefix) :]
                if zero_or_one not in ["1", "0"]:
                    logger.warning(
                        "When using 'done=N', N must be either 0 or 1.",
                        N=zero_or_one,
                    )
                    continue

                done = bool(int(zero_or_one))
                logger.debug(
                    "Filter on whether todo is done or not.", done=done
                )
                continue

            # ----- Metatag Checks
            if word.isalpha():
                metadata_checks.append(MetadataCheck(word))

            if word.startswith("!") and word[1:].isalpha():
                metadata_checks.append(
                    MetadataCheck(
                        word[1:], check=lambda _: False, required=False
                    )
                )

            for op_string, op in [
                ("<=", operator.le),
                (">=", operator.ge),
                ("<", operator.lt),
                (">", operator.gt),
                (":", operator.eq),
            ]:
                key_and_value_string = word.split(op_string)
                if len(key_and_value_string) != 2:
                    continue

                key, value_string = key_and_value_string

                required = not key.endswith("?")
                key = key.rstrip("?")

                value: dt.date | str | int
                if op_string == ":":
                    value = value_string
                elif matches_date_fmt(value_string):
                    value = magodo.to_date(value_string)
                else:
                    value = int(value_string)

                check = _make_metadata_func(op, value)
                metadata_checks.append(
                    MetadataCheck(key, check=check, required=required)
                )
                break

        return cls(
            contexts=contexts,
            create_date=create_and_done[0],
            done_date=create_and_done[1],
            done=done,
            epics=epics,
            metadata_checks=metadata_checks,
            priorities=priorities,
            projects=projects,
        )


def _make_metadata_func(
    op: Callable[[Any, Any], bool], expected: Any
) -> MetadataFunc:
    def check(x: str) -> bool:
        actual: dt.date | str | int
        if isinstance(expected, dt.date):
            actual = magodo.to_date(x)
        elif isinstance(expected, int):
            actual = int(x)
        else:
            actual = x

        return op(actual, expected)

    return check



"""Contains the Repo class."""

from __future__ import annotations

from pathlib import Path

from eris import ErisResult, Ok
from logrus import Logger
from magodo import TodoGroup
from potoroo import TaggedRepo
from typist import PathLike

from ._common import init_yyyymm_path
from ._ids import NULL_ID, init_next_todo_id
from ._tag import Tag
from ._todo import GreatTodo


logger = Logger(__name__)


class GreatRepo(TaggedRepo[str, GreatTodo, Tag]):
    """Repo that stores Todos on disk."""

    def __init__(self, data_dir: PathLike, path: PathLike) -> None:
        self.data_dir = Path(data_dir)
        self.path = Path(path)

    @property
    def todo_group(self) -> TodoGroup[GreatTodo]:
        """Returns the TodoGroup associated with this GreatRepo."""
        return TodoGroup.from_path(GreatTodo, self.path)

    def add(self, todo: GreatTodo, /, *, key: str = None) -> ErisResult[str]:
        """Write a new Todo to disk.

        Returns a unique identifier that has been associated with this Todo.
        """
        drop_old_key = bool(todo.ident == NULL_ID)
        if key is None or key == NULL_ID:
            key = init_next_todo_id(self.data_dir)
        else:
            drop_old_key = True

        mdata = dict(todo.metadata.items())

        old_key = mdata.get("id")
        if (drop_old_key and old_key != key) or not old_key:
            metadata = dict(mdata.items())
            metadata.update({"id": key})
            todo = todo.new(metadata=metadata)

        all_todos: list[GreatTodo] = [todo]

        if self.path.is_dir() or (
            not self.path.exists() and self.path.suffix != ".txt"
        ):
            todo_txt = init_yyyymm_path(self.path, date=todo.create_date)
        else:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            todo_txt = self.path

        if todo_txt.exists():
            todo_group = TodoGroup.from_path(GreatTodo, todo_txt)
            all_todos.extend(todo_group)

        with todo_txt.open("w") as f:
            f.write("\n".join(t.to_line() for t in sorted(all_todos)))

        return Ok(key)

    def get(self, key: str) -> ErisResult[GreatTodo | None]:
        """Retrieve a Todo from disk."""
        return Ok(self.todo_group.todo_map.get(key, None))

    def remove(self, key: str) -> ErisResult[GreatTodo | None]:
        """Remove a Todo from disk."""
        todo_txt = self.todo_group.path_map[key]

        new_lines: list[str] = []

        todo: GreatTodo | None = None
        for line in todo_txt.read_text().split("\n"):
            for word in line.strip().split(" "):
                if word == f"id:{key}":
                    todo = GreatTodo.from_line(line).unwrap()
                    break
            else:
                new_lines.append(line)

        todo_txt.write_text("\n".join(new_lines))

        return Ok(todo)

    def update(self, key: str, todo: GreatTodo, /) -> ErisResult[GreatTodo]:
        """Overwrite an existing Todo on disk."""
        todo_txt = self.todo_group.path_map.get(key)
        if todo_txt is None:
            logger.info(
                "No todo appears to exist with the given key. Adding new todo"
                " instead...",
                key=key,
                todo=todo,
            )
            self.add(todo, key=key)
            return Ok(todo)

        all_lines = []

        old_todo: GreatTodo | None = None
        for line in todo_txt.read_text().split("\n"):
            line = line.strip()
            if not line:
                continue

            if any(w == f"id:{key}" for w in line.split(" ")):
                if todo.to_line() == line:
                    return Ok(todo)

                old_todo = GreatTodo.from_line(line).unwrap()
            else:
                all_lines.append(line)

        if old_todo is None:
            logger.warning(
                "No todo found with this key despite matching todo.txt file?"
                " Adding new todo instead...",
                key=key,
                todo=todo,
            )
            self.add(todo, key=key)
            return Ok(todo)

        all_todos = [GreatTodo.from_line(line).unwrap() for line in all_lines]
        all_todos.append(todo)
        with todo_txt.open("w") as f:
            f.write("\n".join(t.to_line() for t in sorted(all_todos)))

        return Ok(old_todo)

    def get_by_tag(self, tag: Tag) -> ErisResult[list[GreatTodo]]:
        """Get Todos from disk by using a tag.

        Retrieves a list of Todos from disk by using another Todo's properties
        as search criteria.
        """

        return Ok(
            list(
                self.todo_group.filter_by(
                    contexts=tag.contexts,
                    create_date=tag.create_date,
                    done_date=tag.done_date,
                    done=tag.done,
                    epics=tag.epics,
                    metadata_checks=tag.metadata_checks,
                    priorities=tag.priorities,
                    projects=tag.projects,
                )
            )
        )

    def remove_by_tag(self, tag: Tag) -> ErisResult[list[GreatTodo]]:
        """Remove a Todo from disk by using a tag.

        Removes a list of Todos from disk by using another Todo's properties
        as search criteria.
        """
        removed_todos = self.get_by_tag(tag).unwrap()
        for todo in removed_todos:
            self.remove(todo.ident).unwrap()
        return Ok(removed_todos)



"""Contains custom magodo Todo spells used by greatday."""

from __future__ import annotations

import datetime as dt
from functools import partial
from typing import List

from logrus import Logger
import magodo
from magodo.spells import (
    DEFAULT_FROM_LINE_SPELLS,
    DEFAULT_POST_TODO_SPELLS,
    DEFAULT_PRE_TODO_SPELLS,
    DEFAULT_TO_LINE_SPELLS,
    DEFAULT_TODO_SPELLS,
    register_line_spell_factory,
    register_todo_spell_factory,
)
from magodo.types import LineSpell, T, TodoSpell

from ._common import (
    CTX_TODAY,
    drop_word_from_desc,
    dt_from_date_and_hhmm,
    get_relative_date,
    matches_date_fmt,
)


logger = Logger(__name__)

GREAT_PRE_TODO_SPELLS: List[TodoSpell] = list(DEFAULT_PRE_TODO_SPELLS)
pre_todo_spell = register_todo_spell_factory(GREAT_PRE_TODO_SPELLS)

GREAT_TODO_SPELLS: List[TodoSpell] = list(DEFAULT_TODO_SPELLS)
todo_spell = register_todo_spell_factory(GREAT_TODO_SPELLS)

GREAT_POST_TODO_SPELLS: List[TodoSpell] = list(DEFAULT_POST_TODO_SPELLS)
post_todo_spell = register_todo_spell_factory(GREAT_POST_TODO_SPELLS)

GREAT_TO_LINE_SPELLS: List[LineSpell] = list(DEFAULT_TO_LINE_SPELLS)
to_line_spell = register_line_spell_factory(GREAT_TO_LINE_SPELLS)

GREAT_FROM_LINE_SPELLS: List[LineSpell] = list(DEFAULT_FROM_LINE_SPELLS)
from_line_spell = register_line_spell_factory(GREAT_FROM_LINE_SPELLS)


def _startswith_op(x: str, y: str) -> bool:
    """Used as the value for the 'op' kwarg of `drop_word_from_desc()`."""
    return x.startswith(y)


drop_word_if_startswith = partial(drop_word_from_desc, op=_startswith_op)


###############################################################################
# First, all PRE todo spells are cast...
###############################################################################
@pre_todo_spell
def x_points(todo: T) -> T:
    """Handles metatags of the form 'x:N' at the start of a todo line."""
    x = todo.metadata.get("x")
    if not x or len(x) >= 4:
        return todo

    if not todo.desc.startswith("x:"):
        return todo

    metadata = dict(todo.metadata.items())
    points = metadata["x"]
    del metadata["x"]
    metadata["p"] = points

    desc = todo.desc
    desc = drop_word_if_startswith(desc, "p:")
    desc = " ".join(desc.split(" ")[1:]) + f" p:{points}"

    new_todo = todo.new(desc=desc, done=True, metadata=metadata)
    line = new_todo.to_line()
    return type(todo).from_line(line).unwrap()


###############################################################################
# Then all NORMAL todo spells are cast...
###############################################################################
@todo_spell
def snooze_spell(todo: T) -> T:
    """Handles the 'snooze' metadata tag."""
    make_new_todo = False
    metadata = dict(todo.metadata.items())
    s = metadata.get("s")
    if s is not None:
        metadata["snooze"] = s
        del metadata["s"]
        make_new_todo = True

    snooze = metadata.get("snooze")
    if snooze is None:
        return todo

    desc = drop_word_if_startswith(todo.desc, "snooze:")

    today = dt.date.today()

    if matches_date_fmt(snooze):
        snooze_date = magodo.to_date(snooze)
    else:
        snooze_date = get_relative_date(snooze)

    desc = todo.desc

    if snooze_date <= today:
        del metadata["snooze"]
        return todo.new(desc=todo.desc, metadata=metadata)
    elif make_new_todo:
        return todo.new(desc=desc, metadata=metadata)
    else:
        return todo


@todo_spell
def render_tickle_or_snooze_tag(todo: T) -> T:
    """Renders tickler and snooze metatags.

    (e.g. 'tickle:1d' -> 'tickle:2022-02-16')
    """
    found_tag = False
    desc = todo.desc
    metadata = dict(todo.metadata.items())

    for key in ["tickle", "snooze"]:
        # t_or_s: Tickle or Snooze
        t_o_s = todo.metadata.get(key)
        if not t_o_s:
            continue

        if matches_date_fmt(t_o_s):
            continue

        found_tag = True

        new_t_or_s_date = get_relative_date(t_o_s)
        new_t_or_s = magodo.from_date(new_t_or_s_date)
        metadata[key] = new_t_or_s

        desc = drop_word_if_startswith(desc, key + ":")

    if not found_tag:
        return todo

    return todo.new(desc=desc, metadata=metadata)


@todo_spell
def remove_today_context(todo: T) -> T:
    """Removes the @today context from done todos completed before today."""
    if not todo.done_date:
        return todo

    if CTX_TODAY not in todo.contexts:
        return todo

    today = dt.date.today()
    if todo.done_date == today:
        return todo

    contexts = [ctx for ctx in todo.contexts if ctx != CTX_TODAY]
    desc = drop_word_from_desc(todo.desc, f"@{CTX_TODAY}")
    return todo.new(desc=desc, contexts=contexts)


@todo_spell
def recur_spell(todo: T) -> T:
    """Handles the 'recur:' metatag."""
    mdata = todo.metadata

    if not todo.done:
        return todo

    recur = mdata.get("recur")
    if not recur:
        return todo

    tickle = mdata.get("tickle")
    if not tickle:
        return todo

    today = dt.date.today()
    assert isinstance(recur, str)
    if recur.islower():
        start_date = today
    else:
        start_date = magodo.to_date(tickle)

    next_date = get_relative_date(recur, start_date=start_date)
    if next_date <= today:
        logger.warning(
            "Recurring tickler is still due after reset / recur.",
            tickle=next_date,
        )

    metadata = dict(mdata.items())

    if magodo.to_date(tickle) <= today:
        next_tickle_date = next_date
        new_tickle = magodo.from_date(next_tickle_date)
        metadata["tickle"] = new_tickle
    else:
        new_tickle = tickle

    if "dtime" in metadata:
        del metadata["dtime"]

    desc_words = todo.desc.split(" ")
    new_desc_words = []
    for word in desc_words:
        if word.startswith("tickle:"):
            new_desc_words.append(f"tickle:{new_tickle}")
        elif word.startswith("dtime:"):
            continue
        else:
            new_desc_words.append(word)

    desc = " ".join(new_desc_words)

    return todo.new(desc=desc, metadata=metadata, done=False, done_date=None)


@todo_spell
def appt_todos(todo: T) -> T:
    """Adds priority of (C) or (T) to todos with an appt:HHMM tag."""
    appt = todo.metadata.get("appt")
    if not appt:
        return todo

    if todo.done or todo.done_date:
        return todo

    create_date = todo.create_date
    if create_date is None:
        return todo

    now = dt.datetime.now()
    appt_dt = dt_from_date_and_hhmm(create_date, appt)

    if appt_dt < now + dt.timedelta(minutes=30):
        priority = "C"
    else:
        priority = "T"

    return todo.new(priority=priority)


@todo_spell
def i_priority_spell(todo: T) -> T:
    """Handles todos with the in-prigress [i.e. (I)] priority."""
    start = todo.metadata.get("start")
    if todo.priority != "I":
        if start:
            metadata = dict(todo.metadata.items())
            del metadata["start"]
            desc = drop_word_if_startswith(todo.desc, "start:")
            return todo.new(desc=desc, metadata=metadata)
        else:
            return todo

    if start:
        return todo

    now = dt.datetime.now()
    start = f"{now.hour:0>2}{now.minute:0>2}"

    metadata = dict(todo.metadata.items())
    metadata["start"] = start
    desc = todo.desc + f" start:{start}"

    return todo.new(desc=desc, metadata=metadata)


###############################################################################
# Lastly, all POST todo spells are cast...
###############################################################################
@post_todo_spell
def remove_priorities(todo: T) -> T:
    """Remove prioritis for done todos or non-today todos."""
    if todo.priority == magodo.DEFAULT_PRIORITY:
        return todo

    if not todo.done and "today" in todo.contexts:
        return todo

    priority = magodo.DEFAULT_PRIORITY
    desc = drop_word_from_desc(todo.desc, f"({todo.priority})")
    return todo.new(desc=desc, priority=priority)



"""Contains this project's clack runners."""

from __future__ import annotations

from collections import defaultdict
import datetime as dt
import json
from typing import Any, Dict, Final, Iterable, List

import clack
from clack.types import ClackRunner
from logrus import Logger
import magodo

from ._common import CTX_TODAY, drop_word_from_desc, is_tickler
from ._config import AddConfig, InfoConfig, ListConfig, TUIConfig
from ._repo import GreatRepo
from ._session import GreatSession
from ._tag import Tag
from ._todo import GreatTodo
from ._tui import start_textual_app


ALL_RUNNERS: List[ClackRunner] = []
runner = clack.register_runner_factory(ALL_RUNNERS)

logger = Logger(__name__)

CTX_X: Final = "x"
TODO_DIR: Final = "todos"


@runner
def run_add(cfg: AddConfig) -> int:
    """Runner for the 'add' subcommand."""
    log = logger.bind_fargs(locals())

    todo_dir = cfg.data_dir / "todos"
    repo = GreatRepo(cfg.data_dir, todo_dir)
    todo = GreatTodo.from_line(cfg.todo_line).unwrap()

    x_found = False
    if CTX_X in todo.contexts:
        x_found = True
        desc = drop_word_from_desc(todo.desc, f"@{CTX_X}")
        contexts = [ctx for ctx in todo.contexts if ctx != CTX_X]
        todo = todo.new(desc=desc, contexts=contexts)

    if cfg.add_inbox_context == "y" or (
        cfg.add_inbox_context == "default"
        and not x_found
        and not is_tickler(todo)
        and all(ctx not in todo.contexts for ctx in ["inbox", CTX_TODAY])
    ):
        contexts = list(todo.contexts) + ["inbox"]
        todo = todo.new(contexts=contexts)

    key = repo.add(todo).unwrap()
    log.info("Added new todo to inbox.", id=repr(key))
    print(todo.to_line())

    return 0


@runner
def run_info(cfg: InfoConfig) -> int:
    """Runner for the 'info' subcommand."""
    data: Dict[str, Any] = {}

    today = dt.date.today()
    repo_path = cfg.data_dir / TODO_DIR

    stats = data["stats"] = {}
    points_data = stats["points"] = {}
    day_info = points_data["by_day"] = defaultdict(dict)

    # e.g. 'stats.count.total' OR 'stats.count.open' OR 'stats.count.done'
    counter = stats["count"] = {}

    repo = GreatRepo(cfg.data_dir, repo_path)

    # 'stats.count' counter values
    done_count = 0
    open_count = 0
    tickler_count = 0

    for todo in repo.todo_group:
        if todo.done:
            done_count += 1
        else:
            open_count += 1

        # --- loop variables
        # key: used to index into the 'counter' dict.
        # tags: a dict of tags (e.g. projects) from the current todo.
        for key, tags in [
            ("epic", todo.epics),
            ("project", todo.projects),
            ("context", todo.contexts),
        ]:
            for tag in tags:
                if key not in counter:
                    counter[key] = defaultdict(int)

                counter[key][tag] += 1

        if "tickle" in list(todo.metadata.keys()):
            tickler_count += 1

    counter["done"] = done_count
    counter["open"] = open_count
    counter["tickler"] = tickler_count
    counter["total"] = open_count + done_count

    for days in range(cfg.points_start_offset, cfg.points_end_offset + 1):
        # inner (i.e. local to this for-loop) date
        date = today - dt.timedelta(days=days)

        day_total = 0
        xp_day_total = 0

        # Maps '@' contexts to numeric points.
        ctx_to_points: dict[str, int] = defaultdict(int)

        is_today = bool(date == today)

        with GreatSession(
            cfg.data_dir,
            repo_path,
            Tag(
                done_date=date,
                done=True,
                metadata_checks=[magodo.MetadataCheck("p")],
            ),
            name="info",
        ) as done_session:
            for todo in done_session.repo.todo_group:
                P = int(todo.metadata.get("p", 0))
                day_total += P
                xp_day_total += P
                for ctx in todo.contexts:
                    ctx_to_points[ctx] += P

        if is_today:
            # Add sum of 'xp' metatag values for todos due today...
            with GreatSession(
                cfg.data_dir,
                repo_path,
                Tag(
                    done=False,
                    contexts=["today"],
                    metadata_checks=[magodo.MetadataCheck("xp")],
                ),
            ) as open_session:
                for todo in open_session.repo.todo_group:
                    XP = int(todo.metadata.get("xp", 0))
                    xp_day_total += XP

        if day_total or (xp_day_total and is_today):
            date_str = magodo.from_date(date)
            # Only show 'xp_total' for today's date...
            if is_today:
                day_info[date_str]["xp_total"] = xp_day_total

            day_info[date_str]["total"] = day_total
            day_info[date_str]["contexts"] = ctx_to_points

    pretty_data = json.dumps(data, indent=2, sort_keys=True)
    print(pretty_data)

    return 0


@runner
def run_list(cfg: ListConfig) -> int:
    """Runner for the 'list' subcommand."""
    repo_path = cfg.data_dir / TODO_DIR
    repo = GreatRepo(cfg.data_dir, repo_path)

    todo_iter: Iterable[GreatTodo]
    if cfg.query is None:
        todo_iter = repo.todo_group
    else:
        tag = Tag.from_query(cfg.query)
        todo_iter = repo.get_by_tag(tag).unwrap()

    for todo in sorted(todo_iter):
        print(todo.to_line())

    return 0


@runner
def run_tui(cfg: TUIConfig) -> int:
    """Runer for the 'tui' subcommand."""
    repo_path = cfg.data_dir / TODO_DIR
    start_textual_app(cfg.data_dir, repo_path)
    return 0



"""Contains this project's clack.Config classes."""

from __future__ import annotations

import itertools as it
from pathlib import Path
from typing import Any, Literal, Optional, Sequence

import clack
from clack import xdg

from . import APP_NAME
from .types import YesNoDefault


Command = Literal["add", "info", "list", "tui"]


class Config(clack.Config):
    """Shared clack configuration class."""

    command: Command

    # ----- CONFIG
    data_dir: Path = xdg.get_full_dir("data", APP_NAME)


class AddConfig(Config):
    """Config for the 'add' subcommand."""

    command: Literal["add"]

    # ----- ARGUMENTS
    todo_line: str

    # ----- CONFIG
    add_inbox_context: YesNoDefault = "default"


class InfoConfig(Config):
    """Config for the 'info' subcommand."""

    command: Literal["info"]

    # ----- CONFIG
    points_start_offset: int = 0
    points_end_offset: int = 4


class ListConfig(Config):
    """Config for the 'list' subcommand."""

    command: Literal["list"]

    # ----- ARGUMENTS
    query: Optional[str] = None


class TUIConfig(Config):
    """Config for the 'tui' subcommand."""

    command: Literal["tui"]


def clack_parser(argv: Sequence[str]) -> dict[str, Any]:
    """Parser we pass to the `main_factory()` `parser` kwarg."""
    # HACK: Make 'tui' the default sub-command.
    if not list(it.dropwhile(lambda x: x.startswith("-"), argv[1:])):
        argv = list(argv) + ["tui"]

    parser = clack.Parser(
        description="Don't have a good day. Have a great day."
    )

    new_command = clack.new_command_factory(parser)

    # ----- 'tui' command (default)
    new_command(
        "tui",
        help=(
            "Render greatday's text-based user interface (TUI). This is the"
            " default command."
        ),
    )

    # ----- 'add' command
    add_parser = new_command("add", help="Add a new todo to your inbox.")
    add_parser.add_argument(
        "todo_line",
        metavar="TODO",
        help=(
            "A valid todo string (i.e. a string that conforms to the standard"
            " todo.txt format)."
        ),
    )

    # ----- 'info' command
    new_command(
        "info",
        help=(
            "Print information about greatday and its current state to stdout"
            " in JSON format."
        ),
    )

    # ----- 'list' command
    list_parser = new_command("list", help="Query the todo database.")
    list_parser.add_argument(
        "query",
        nargs="?",
        help=(
            "The todo search query that will be used to filter todos. If not"
            " provided, all todos are selected."
        ),
    )

    args = parser.parse_args(argv[1:])
    kwargs = clack.filter_cli_args(args)

    return kwargs



"""Contains greatday's custom Todo types."""

from __future__ import annotations

from magodo import MagicTodoMixin

from . import _spells as spells
from ._ids import NULL_ID


class GreatTodo(MagicTodoMixin):
    """Custom MagicTodo type used when working with Todos in the daily file."""

    pre_todo_spells = spells.GREAT_PRE_TODO_SPELLS
    todo_spells = spells.GREAT_TODO_SPELLS
    post_todo_spells = spells.GREAT_POST_TODO_SPELLS

    to_line_spells = spells.GREAT_TO_LINE_SPELLS
    from_line_spells = spells.GREAT_FROM_LINE_SPELLS

    @property
    def ident(self) -> str:
        """Returns this Todo's unique identifier."""
        result = self.metadata.get("id", NULL_ID)
        assert isinstance(result, str)
        return result



"""Contains the GreatSession class."""

from __future__ import annotations

import os
from pathlib import Path
import tempfile
from types import TracebackType
from typing import Type

from logrus import Logger
from potoroo import UnitOfWork
from typist import PathLike

from ._ids import NULL_ID
from ._repo import GreatRepo
from ._tag import Tag


logger = Logger(__name__)


class GreatSession(UnitOfWork[GreatRepo]):
    """Each time todos are opened in an editor, a new session is created."""

    def __init__(
        self,
        data_dir: PathLike,
        repo_path: PathLike,
        tag: Tag,
        *,
        name: str = None,
    ) -> None:
        self.data_dir = Path(data_dir)
        self._master_path = Path(repo_path)

        prefix = None if name is None else f"{name}."
        _, temp_path = tempfile.mkstemp(prefix=prefix, suffix=".txt")
        self.path = Path(temp_path)

        self._temp_repo = GreatRepo(self.data_dir, self.path)

        self._master_repo = GreatRepo(self.data_dir, self._master_path)
        for todo in self._master_repo.get_by_tag(tag).unwrap():
            self.repo.add(todo, key=todo.ident)

        self._old_todos = list(self._temp_repo.todo_group)

    def __enter__(self) -> GreatSession:
        """Called before entering a GreatSession with-block."""
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Called before exiting a GreatSession with-block."""
        del exc_type
        del exc_value
        del traceback

        os.unlink(self.path)

    def commit(self) -> None:
        """Commit our changes.

        We achieve this by copying the contents of the backup file created on
        instantiation back to the original.
        """
        old_todo_keys = [todo.ident for todo in self._old_todos]
        new_todos = {}
        for todo in self.repo.todo_group:
            key = todo.ident
            if key == NULL_ID:
                logger.info("New todo was added while editing?", todo=todo)
                key = self._master_repo.add(todo).unwrap()
                new_todos[key] = todo
            else:
                self._master_repo.update(key, todo).unwrap()
                if key in old_todo_keys:
                    old_todo_keys.remove(key)

        if new_todos:
            old_lines = self.path.read_text().split("\n")
            self.path.write_text(
                "\n".join(line for line in old_lines if " id:" in line)
            )

        for key, todo in new_todos.items():
            self.repo.add(todo, key=key)

        for key in old_todo_keys:
            removed_todo = self._master_repo.remove(key).unwrap()
            if removed_todo is not None:
                self._old_todos.remove(removed_todo)

    def rollback(self) -> None:
        """Revert any changes made while in this GreatSession's with-block."""

    @property
    def repo(self) -> GreatRepo:
        """Returns the GreatRepo object associated with this GreatSession."""
        return self._temp_repo



"""ID generating logic lives here."""

from __future__ import annotations

from pathlib import Path
from typing import Container, Final

from typist import PathLike


NULL_ID: Final = "null"


def init_next_todo_id(data_dir: PathLike) -> str:
    """Retrieves the next valid todo ID.

    Side Effects:
        * Attempts to read last ID from disk.
        * Writes the returned ID to disk.
    """
    data_dir = Path(data_dir)
    last_id_path = data_dir / "last_todo_id"
    last_id_path.parent.mkdir(parents=True, exist_ok=True)

    def ID(next_id: str) -> str:
        with last_id_path.open("w+"):
            last_id_path.write_text(next_id)
        return next_id

    if last_id_path.exists():
        last_id = last_id_path.read_text().strip()
        next_id = next_todo_id(last_id)
        return ID(next_id)
    else:
        return ID("0")


def next_todo_id(last_id: str) -> str:
    """Determines the next ID from the last ID.

    Examples:
        >>> next_todo_id('0')
        '1'

        >>> next_todo_id('9')
        'A'

        >>> next_todo_id('Z')
        '00'

        >>> next_todo_id('ZZ')
        '000'

        >>> next_todo_id('AZ')
        'B0'

        >>> next_todo_id('BM9')
        'BMA'

        >>> next_todo_id('BMZ')
        'BN0'

        >>> next_todo_id('BZZ')
        'C00'

        >>> next_todo_id('0ZZ')
        '100'

        >>> next_todo_id('C00')
        'C01'

        # we skip 'I', since it can be confused with '1'...
        >>> next_todo_id('BZH')
        'BZJ'

        # we skip 'O', since it can be confused with '0'...
        >>> next_todo_id('BZN')
        'BZP'

        >>> next_todo_id('B9Z')
        'BA0'
    """
    for i, ch in enumerate(reversed(last_id)):
        if ch != "Z":
            idx = len(last_id) - (i + 1)
            zeros = "0" * i
            return last_id[:idx] + next_char(last_id[idx]) + zeros

    zeros = "0" * (len(last_id) + 1)
    return zeros


def next_char(ch: str, *, blacklist: Container[str] = ("I", "O")) -> str:
    """Returns the next allowable character (to be used as apart of ID)."""
    assert len(ch) == 1, (
        "PRE-CONDITION: The `ch` argument must be a single-character string."
        f" | ch={ch!r}"
    )

    if ch == "9":
        return "A"

    result = chr(ord(ch) + 1)
    while result in blacklist:
        result = chr(ord(result) + 1)
    return result
